# -*- coding: utf-8 -*-
"""Init and utils."""
from zope.i18nmessageid import MessageFactory

import logging


_ = MessageFactory("collective.timestamp")
logger = logging.getLogger("collective.timestamp")



# -*- coding: utf-8 -*-

from collective.timestamp import _
from collective.timestamp.interfaces import ITimeStamper
from plone import api


def modified_content(obj, event):
    handler = ITimeStamper(obj)
    if not handler.is_timestamped():
        # object is not timestamped, nothing to do here
        return
    if not handler.file_has_changed(obj, event):
        # primary file didn't change, nothing to do
        return
    # primary file field has changed, we need to remove timestamp
    obj.timestamp = None
    obj.reindexObject(idxs=["is_timestamped"])
    request = getattr(obj, "REQUEST", None)
    if request is not None:
        message = _("Timestamp information has been removed since the data has changed")
        api.portal.show_message(message, request)


def warn_on_edition(obj, event):
    handler = ITimeStamper(obj)
    if not handler.is_timestamped():
        # object is not timestamped, no need to warn the user
        return
    request = getattr(obj, "REQUEST", None)
    if request is not None:
        message = _(
            "You are editing a timestamped content. Your modifications can invalidate the timestamp."
        )
        api.portal.show_message(message, request, type="warning")



from collective.timestamp.interfaces import ITimeStamper
from collective.timestamp.testing import COLLECTIVE_TIMESTAMP_FUNCTIONAL_TESTING
from collective.timestamp.testing import COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING
from pathlib import Path
from plone import api
from plone.api.portal import set_registry_record
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.namedfile import NamedBlobFile
from pyasn1.codec.der import decoder
from rfc3161ng import TimeStampResp
from unittest.mock import patch

import unittest


class TimestampBaseTestCase(unittest.TestCase):
    """Base test case for the collective.timestamp package."""

    def setUp(self):
        """Set up the test case."""
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.document = api.content.create(
            container=self.portal,
            type="Document",
            id="my-document",
        )
        self.file = api.content.create(
            container=self.portal,
            type="File",
            id="my-file",
        )
        self.timestamped_file = api.content.create(
            container=self.portal,
            type="File",
            id="timestamped-file",
        )
        self.timestamped_file.file = NamedBlobFile(
            data=b"file data", filename="file.txt"
        )
        mock_tsr_path = Path(__file__).parent / "resources/mock_tsr_file.tsr"
        with open(mock_tsr_path, "rb") as f:
            self.raw_tsr = f.read()
        self.tsr, _ = decoder.decode(self.raw_tsr, asn1Spec=TimeStampResp())
        with patch("rfc3161ng.api.RemoteTimestamper.__call__") as mock_timestamper:
            mock_timestamper.return_value = self.tsr
            handler = ITimeStamper(self.timestamped_file)
            handler.timestamp()
        logout()


class TimestampIntegrationTestCase(TimestampBaseTestCase):
    """Integration test case for the collective.timestamp package."""

    layer = COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]


class TimestampFunctionalTestCase(TimestampBaseTestCase):
    """Functional test case for the collective.timestamp package."""

    layer = COLLECTIVE_TIMESTAMP_FUNCTIONAL_TESTING

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]



# -*- coding: utf-8 -*-

from collective.timestamp.testing import COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.CMFPlone.utils import get_installer

import unittest


class TestSetup(unittest.TestCase):
    """Test that collective.timestamp is properly installed."""

    layer = COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.portal = self.layer["portal"]
        self.installer = get_installer(self.portal, self.layer["request"])

    def test_product_installed(self):
        """Test if collective.timestamp is installed."""
        self.assertTrue(self.installer.is_product_installed("collective.timestamp"))

    def test_browserlayer(self):
        """Test that ICollectiveTimestampLayer is registered."""
        from collective.timestamp.interfaces import ICollectiveTimestampLayer
        from plone.browserlayer import utils

        self.assertIn(ICollectiveTimestampLayer, utils.registered_layers())


class TestUninstall(unittest.TestCase):

    layer = COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.installer = get_installer(self.portal, self.layer["request"])
        roles_before = api.user.get_roles(TEST_USER_ID)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.installer.uninstall_product("collective.timestamp")
        setRoles(self.portal, TEST_USER_ID, roles_before)

    def test_product_uninstalled(self):
        """Test if collective.timestamp is cleanly uninstalled."""
        self.assertFalse(self.installer.is_product_installed("collective.timestamp"))

    def test_browserlayer_removed(self):
        """Test that ICollectiveTimestampLayer is removed."""
        from collective.timestamp.interfaces import ICollectiveTimestampLayer
        from plone.browserlayer import utils

        self.assertNotIn(ICollectiveTimestampLayer, utils.registered_layers())



# -*- coding: utf-8 -*-
from plone.app.robotframework.testing import REMOTE_LIBRARY_BUNDLE_FIXTURE
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.testing import z2

import collective.timestamp


class CollectiveTimestampLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # Load any other ZCML that is required for your tests.
        # The z3c.autoinclude feature is disabled in the Plone fixture base
        # layer.
        import plone.app.dexterity

        self.loadZCML(package=plone.app.dexterity)
        import plone.restapi

        self.loadZCML(package=plone.restapi)
        self.loadZCML(package=collective.timestamp)

    def setUpPloneSite(self, portal):
        applyProfile(portal, "collective.timestamp:testing")


COLLECTIVE_TIMESTAMP_FIXTURE = CollectiveTimestampLayer()


COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING = IntegrationTesting(
    bases=(COLLECTIVE_TIMESTAMP_FIXTURE,),
    name="CollectiveTimestampLayer:IntegrationTesting",
)


COLLECTIVE_TIMESTAMP_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(COLLECTIVE_TIMESTAMP_FIXTURE,),
    name="CollectiveTimestampLayer:FunctionalTesting",
)


COLLECTIVE_TIMESTAMP_ACCEPTANCE_TESTING = FunctionalTesting(
    bases=(
        COLLECTIVE_TIMESTAMP_FIXTURE,
        REMOTE_LIBRARY_BUNDLE_FIXTURE,
        z2.ZSERVER_FIXTURE,
    ),
    name="CollectiveTimestampLayer:AcceptanceTesting",
)



# -*- coding: utf-8 -*-

from collective.timestamp import logger
from collective.timestamp.interfaces import ITimeStamper
from collective.timestamp.interfaces import ITimestampingSettings
from collective.timestamp.utils import timestamp
from plone.namedfile.file import NamedBlobFile
from plone.namedfile.interfaces import INamedField
from plone.registry.interfaces import IRegistry
from plone.rfc822.interfaces import IPrimaryFieldInfo
from zope.component import getUtility
from zope.interface import implementer
from zope.lifecycleevent.interfaces import IAttributes


@implementer(ITimeStamper)
class TimeStamper(object):
    """Handle timestamping operations on an object"""

    def __init__(self, context):
        self.context = context

    def get_file_field(self):
        try:
            primary = IPrimaryFieldInfo(self.context, None)
            if (
                INamedField.providedBy(primary.field)
                and hasattr(primary.value, "getSize")
                and primary.value.getSize() > 0
            ):
                return primary
        except TypeError:
            pass

    def get_data(self):
        field = self.get_file_field()
        if field is None:
            logger.warning(
                f"Could not find the file field for {self.context.absolute_url()}"
            )
            return
        return field.value.data

    def file_has_changed(self, obj, event):
        field = self.get_file_field()
        fieldname = field.fieldname
        for d in event.descriptions:
            if not IAttributes.providedBy(d):
                continue
            if fieldname in d.attributes:
                return True
        return False

    def is_timestamped(self):
        return self.context.timestamp is not None

    def is_timestampable(self):
        if not self.context.enable_timestamping:
            return False
        elif self.is_timestamped():
            return False
        return self.get_data() is not None

    def _effective_related_indexes(self):
        return ["effective", "effectiveRange", "is_timestamped"]

    def generate_timestamp(self, file_content: bytes):
        """Produce the .tsr blob and the datetime when it was created, using the registry settings."""
        settings = getUtility(IRegistry).forInterface(ITimestampingSettings)
        return timestamp(
            file_content,
            service_url=settings.timestamping_service_url,
            hashing_algorithm=settings.hashing_algorithm,
            use_failover=settings.use_failover,
            failover_timestamping_service_urls=settings.failover_timestamping_service_urls,
            max_retries=settings.max_retries,
            initial_backoff_seconds=settings.initial_backoff_seconds,
        )

    def timestamp(self):
        """Timestamp this context: attach the .tsr file, set the effective date, and reindex."""
        if not self.is_timestampable():
            raise ValueError("This content is not timestampable")
        data = self.get_data()
        tsr_file, timestamp_date = self.generate_timestamp(data)
        self.context.timestamp = NamedBlobFile(data=tsr_file, filename="timestamp.tsr")
        self.context.setEffectiveDate(timestamp_date)
        self.context.reindexObject(idxs=self._effective_related_indexes())
        # return data and timestamp in case method is overrided
        return data, timestamp_date



# -*- coding: utf-8 -*-

from asn1crypto import cms
from asn1crypto import tsp
from asn1crypto import x509
from collective.timestamp import _
from collective.timestamp.interfaces import ITimeStamper
from collective.timestamp.interfaces import ITimestampingSettings
from collective.timestamp.utils import get_timestamp_date_from_tsr_file
from plone import api
from Products.Five.browser import BrowserView
from zope.i18n import translate

import hashlib


RDN_KEY_I18N_MAP = {
    "common_name": _("label_common_name", default="Common Name"),
    "organization_identifier": _(
        "label_organization_identifier", default="Organization Identifier"
    ),
    "organization_name": _("label_organization_name", default="Organization"),
    "organizational_unit_name": _(
        "label_organizational_unit_name", default="Org. Unit"
    ),
    "locality_name": _("label_locality_name", default="Locality"),
    "state_or_province_name": _(
        "label_state_province_name", default="State / Province"
    ),
    "country_name": _("label_country_name", default="Country"),
    "serial_number": _("label_serial_number", default="Serial Number"),
    "email_address": _("label_email_address", default="E-mail"),
    "domain_component": _("label_domain_component", default="Domain"),
}


class TimestampInfo(BrowserView):

    def __init__(self, context, request):
        super().__init__(context, request)
        self.tst_info, self.token_ci, self.signed_data = self._load_tst_info()

    def _load_tst_info(self):
        """
        Return a tsp.TSTInfo instance if the timestamp was granted,
        otherwise None.
        """
        if self.context.timestamp is None:
            return None, None, None
        tsr = self.context.timestamp
        data = tsr.data

        # Decode the outer TimeStampResp
        resp = tsp.TimeStampResp.load(data)
        if resp["status"]["status"].native not in ("granted", "granted_with_mods"):
            return None, None, None

        # Walk down into the CMS Time-Stamp Token
        token_ci = cms.ContentInfo.load(resp["time_stamp_token"].dump())
        signed = token_ci["content"]  # cms.SignedData
        eci = signed["encap_content_info"]  # cms.EncapContentInfo
        content_os = eci["content"]  # OctetString (maybe parsed)

        # If asn1crypto has already parsed the octet-string, use it, otherwise fall back to manual load from bytes.
        try:
            tst_info = content_os.parsed  # preferred path
        except (ValueError, AttributeError):
            raw = content_os.native
            if not isinstance(raw, (bytes, bytearray)):
                raw = content_os.dump()
            tst_info = tsp.TSTInfo.load(raw)

        return tst_info, token_ci, signed  # Return TSTInfo, ContentInfo, SignedData

    def is_timestamped(self):
        handler = ITimeStamper(self.context)
        return handler.is_timestamped()

    def timestamp_date(self):
        tsr_file = self.context.timestamp
        timestamp_date = get_timestamp_date_from_tsr_file(tsr_file.data)
        return timestamp_date.astimezone()

    def timestamp_file_link(self):
        """Return dict {url, label} for the timestamped file."""
        handler = ITimeStamper(self.context)
        field = handler.get_file_field()
        if not field:
            return None
        fieldname = field.fieldname
        value = getattr(self.context, fieldname, None)
        if not value:
            return None
        filename = getattr(value, "filename", None) or "download"
        return {
            "url": f"{self.context.absolute_url()}/@@download/{fieldname}/{filename}",
            "label": filename,
        }

    def timestamp_tsr_link(self):
        """Return dict {url, label} for the TSR file."""
        tsr = getattr(self.context, "timestamp", None)
        if not tsr:
            return None
        filename = getattr(tsr, "filename", None) or "timestamp.tsr"
        return {
            "url": f"{self.context.absolute_url()}/@@download/timestamp/{filename}",
            "label": filename,
        }

    def more_infos_url(self):
        return api.portal.get_registry_record(
            "timestamping_documentation_url",
            interface=ITimestampingSettings,
        )

    def timestamp_policy_oid(self):
        """Policy OID (string, e.g., '1.3.6.1.4.1.…')."""
        if self.tst_info is None:
            return None
        return self.tst_info["policy"].dotted

    def timestamp_serial_number(self):
        """Serial number (int)."""
        if self.tst_info is None:
            return None
        return self.tst_info["serial_number"].native

    def encap_content_type_oid(self):
        """
        OID of the EncapsulatedContentInfo content-type.
        For RFC 3161 tokens this is typically id-ct-TSTInfo = 1.2.840.113549.1.9.16.1.4
        """
        if self.context.timestamp is None:
            return None
        resp = tsp.TimeStampResp.load(self.context.timestamp.data)
        token_ci = cms.ContentInfo.load(resp["time_stamp_token"].dump())
        return token_ci["content"]["encap_content_info"]["content_type"].dotted

    def timestamp_authority(self):
        """Returns a dict with the TSA authority information."""
        if self.tst_info is None:
            return {}

        tsa = self.tst_info["tsa"]

        # Fallback in case the TSA is not a directory name
        if tsa.name != "directory_name":
            req = getattr(self, "request", None)
            return {
                "general_name": {
                    "label": translate(
                        _("label_general_name", default="General Name"),
                        context=req,
                    ),
                    "value": str(tsa.chosen.native),
                }
            }

        name: x509.Name = tsa.chosen

        info = {}
        for rdn in name.chosen:
            for ava in rdn:
                oid = ava["type"].native
                value = ava["value"].native
                label = RDN_KEY_I18N_MAP.get(oid, oid)
                info[oid] = {"label": label, "value": str(value)}

        return info

    def timestamp_precision(self):
        """
        Returns the accuracy of the timestamp in a human-friendly string.

        Examples:
        "1.234 second(s)"
        "0.5 second(s)"
        "0 second (exact)"
        "Not specified"
        """

        if self.tst_info is None:
            return None
        accuracy = self.tst_info.native.get("accuracy", None)
        if not accuracy:
            return _("Not specified")

        seconds = accuracy.get("seconds", 0) or 0
        millis = accuracy.get("millis", 0) or 0
        micros = accuracy.get("micros", 0) or 0

        total_seconds = seconds + millis / 1_000 + micros / 1_000_000

        if total_seconds == 0:
            return _("0 second (exact)")

        if micros:
            fmt = "{:.6f}"
        elif millis:
            fmt = "{:.3f}"
        else:
            fmt = "{:.0f}"

        human = fmt.format(total_seconds).rstrip("0").rstrip(".")
        return _("{} second(s)").format(human)

    def timestamp_protocol(self):
        return "RFC 3161" if self.tst_info is not None else None

    def timestamp_hash(self):
        """
        The message imprint digest (hex-encoded)
        """
        if self.tst_info is None:
            return None
        digest = self.tst_info["message_imprint"]["hashed_message"].native
        # native is bytes
        return digest.hex()

    def timestamp_algorithm(self):
        """
        The hash algorithm used in the message imprint (e.g. 'sha256')
        """
        if self.tst_info is None:
            return None
        alg = self.tst_info["message_imprint"]["hash_algorithm"]["algorithm"].native
        return alg

    def signer_certificate(self):
        """
        Return the asn1crypto.x509.Certificate of the token signer, or None.
        # TODO: https://github.com/wbond/oscrypto has some utilities for this.
        """
        signed = self.signed_data
        if signed is None:
            return None

        try:
            signer_infos = signed["signer_infos"]
            if len(signer_infos) == 0:
                return None
            sid = signer_infos[0]["sid"]
        except Exception:
            return None

        certs_field = signed["certificates"] if "certificates" in signed else []
        if not certs_field:
            return None

        # Match by IssuerAndSerialNumber (most common)
        if sid.name == "issuer_and_serial_number":
            iasn = sid.chosen
            sid_issuer = iasn["issuer"]  # x509.Name
            sid_serial = iasn["serial_number"].native
            for cert_choice in certs_field:
                if cert_choice.name != "certificate":
                    continue
                cert = cert_choice.chosen  # Unwrap CertificateChoices
                tbs = cert["tbs_certificate"]
                if tbs["serial_number"].native != sid_serial:
                    continue
                # Compare issuer by DER to be robust
                if tbs["issuer"].dump() == sid_issuer.dump():
                    return cert
            return None

        # Match by SubjectKeyIdentifier if present
        if sid.name == "subject_key_identifier":
            ski = sid.chosen.native  # bytes
            for cert_choice in certs_field:
                if cert_choice.name != "certificate":
                    continue
                cert = cert_choice.chosen
                try:
                    for ext in cert["tbs_certificate"]["extensions"]:
                        if ext["extn_id"].native == "subject_key_identifier":
                            if ext["extn_value"].native == ski:
                                return cert
                except Exception:
                    continue
            return None
        return None

    def timestamp_signer_subject(self):
        """
        Returns a dict mapping attribute OIDs (e.g., 'common_name') to:
          {'label': <translated label>, 'value': <string>}
        for the signer certificate's *subject*. Empty dict if unavailable.
        """
        cert = self.signer_certificate()
        if cert is None:
            return {}
        subject: x509.Name = cert["tbs_certificate"]["subject"]

        info = {}
        for rdn in subject.chosen:
            for ava in rdn:
                oid = ava["type"].native
                value = ava["value"].native
                label = RDN_KEY_I18N_MAP.get(oid, oid)
                info[oid] = {"label": label, "value": str(value)}
        return info

    @staticmethod
    def _get_ext_value(cert, extn_id):
        """
        Return extension .native value for extn_id or None.
        extn_id examples: 'subject_key_identifier', 'authority_key_identifier'
        """
        try:
            for ext in cert["tbs_certificate"]["extensions"]:
                if ext["extn_id"].native == extn_id:
                    return ext["extn_value"].native
        except Exception:
            return None
        return None

    @classmethod
    def _get_ski(cls, cert):
        """
        Subject Key Identifier bytes, or None if not present.
        """
        val = cls._get_ext_value(cert, "subject_key_identifier")
        return None if val is None else bytes(val)

    @classmethod
    def _get_aki(cls, cert):
        """
        Authority Key Identifier dict (may contain 'key_identifier', 'authority_cert_issuer',
        'authority_cert_serial_number'), or None if not present.
        """
        val = cls._get_ext_value(cert, "authority_key_identifier")
        if val is None:
            return None
        # Normalize key_identifier to bytes if present
        kid = val.get("key_identifier")
        if kid is not None:
            val = dict(val)
            val["key_identifier"] = bytes(kid)
        return val

    @staticmethod
    def _is_self_issued(cert):
        """
        Heuristic: subject == issuer (DER equality).
        Not a cryptographic verification.
        """
        tbs = cert["tbs_certificate"]
        return tbs["subject"].dump() == tbs["issuer"].dump()

    def certificates_in_token(self):
        """
        Return a list[x509.Certificate] of all certificates present in the token.
        TODO: https://github.com/wbond/certvalidator has some utilities for this and does proper validation.
        """
        signed = self.signed_data
        if signed is None:
            return []
        if "certificates" not in signed:
            return []
        certs = signed[
            "certificates"
        ]  # cms.CertificateSet (SequenceOf CertificateChoices)
        return [choice.chosen for choice in certs if choice.name == "certificate"]

    def certificate_chain(self):
        """
        Build a best-effort ordered certificate chain from the signer up to
        the highest ancestor found *within the token's certificate bag*.

        Returns: list[x509.Certificate], where element 0 is the signer.
        Does not do any external fetching or trust evaluation.
        """
        signer = self.signer_certificate()
        if signer is None:
            return []

        all_certs = self.certificates_in_token()
        if not all_certs:
            return [signer]

        # Map SKI -> cert, subjectDN DER -> cert, and (issuerDN DER, serial) -> cert
        ski_map = {}
        subj_map = {}
        serial_map = {}  # (issuer_der, serial_int) -> cert
        for c in all_certs:
            tbs = c["tbs_certificate"]
            subj_der = tbs["subject"].dump()
            iss_der = tbs["issuer"].dump()
            serial = tbs["serial_number"].native
            subj_map[subj_der] = c
            serial_map[(iss_der, serial)] = c
            ski = self._get_ski(c)
            if ski is not None:
                ski_map[ski] = c

        chain = [signer]
        seen = {signer.dump()}

        # Walk upwards using AKI -> SKI if present, else issuer DN (and serial if AKI specifies it)
        current = signer
        for _ in range(len(all_certs)):  # prevent cycles
            tbs = current["tbs_certificate"]
            if self._is_self_issued(current):
                break  # reached (likely) trust anchor within the bag

            aki = self._get_aki(current)
            parent = None

            if aki and aki.get("key_identifier") is not None:
                parent = ski_map.get(aki["key_identifier"])

            if parent is None:
                issuer_der = tbs["issuer"].dump()
                if aki and aki.get("authority_cert_serial_number") is not None:
                    parent = serial_map.get(
                        (issuer_der, aki["authority_cert_serial_number"])
                    )
                if parent is None:
                    parent = subj_map.get(issuer_der)

            if parent is None or parent.dump() in seen:
                break

            chain.append(parent)
            seen.add(parent.dump())
            current = parent

        return chain

    def _name_to_dict(self, name_obj):
        """
        Convert an asn1crypto.x509.Name into a dict similar to timestamp_authority().
        """
        info = {}
        for rdn in name_obj.chosen:
            for ava in rdn:
                oid = ava["type"].native
                value = ava["value"].native
                label = RDN_KEY_I18N_MAP.get(oid, oid)
                info[oid] = {"label": label, "value": str(value)}
        return info

    def certificate_chain_info(self, fingerprint_algorithm="sha256"):
        """
        Return a list of dicts describing the chain.
        Each item has:
          - subject: dict (as in timestamp_authority())
          - issuer: dict
          - serial_number: int
          - fingerprint: hex string (DER over algorithm)
          - self_issued: bool
          - has_ski: bool
          - has_aki: bool
        """
        items = []
        for cert in self.certificate_chain():
            tbs = cert["tbs_certificate"]
            subject = tbs["subject"]
            issuer = tbs["issuer"]
            validity = tbs["validity"]
            fp = hashlib.new(fingerprint_algorithm, cert.dump()).hexdigest()
            not_before = validity["not_before"].native
            not_after = validity["not_after"].native
            items.append(
                {
                    "subject": self._name_to_dict(subject),
                    "issuer": self._name_to_dict(issuer),
                    "serial_number": tbs["serial_number"].native,
                    "fingerprint": fp,
                    "self_issued": self._is_self_issued(cert),
                    "has_ski": self._get_ski(cert) is not None,
                    "has_aki": self._get_aki(cert) is not None,
                    "not_before": not_before,
                    "not_after": not_after,
                }
            )
        return items



# -*- coding: utf-8 -*-
"""Module where all interfaces, events and exceptions live."""

from collective.timestamp import _
from zope import schema
from zope.interface import Interface
from zope.publisher.interfaces.browser import IDefaultBrowserLayer


class ICollectiveTimestampLayer(IDefaultBrowserLayer):
    """Marker interface that defines a browser layer."""


class ITimeStamper(Interface):
    """"""


class ITimestampingSettings(Interface):

    timestamping_service_url = schema.URI(
        title=_("URL of the timestamping service you want to use"),
        default="http://freetsa.org/tsr",
        required=True,
    )

    hashing_algorithm = schema.Choice(
        title=_("Hashing algorithm to use for timestamping"),
        description=_(
            "Choose the hashing algorithm that will be used for generating timestamps."
        ),
        values=["sha1", "sha256", "sha384", "sha512"],
        default="sha256",
        required=True,
    )

    max_retries = schema.Int(
        title=_("Maximum number of retries for timestamping"),
        description=_(
            "The maximum number of attempts to retry timestamping in case of failure."
        ),
        default=0,
        required=True,
    )

    initial_backoff_seconds = schema.Float(
        title=_("Initial backoff time in seconds"),
        description=_(
            "The initial time to wait before retrying a failed timestamping attempt."
        ),
        default=0.5,
        required=True,
    )

    use_failover = schema.Bool(
        title=_("Use failover timestamping"),
        description=_(
            "If enabled, additional timestamping services will be used if the main service is unavailable."
        ),
        default=False,
        required=False,
    )

    failover_timestamping_service_urls = schema.List(
        title=_("List of failover timestamping service URLs"),
        description=_(
            "These services will be used if the main service is unavailable."
        ),
        value_type=schema.URI(),
        default=[],
        required=False,
    )

    timestamping_documentation_url = schema.URI(
        title=_("URL of the documentation explaining how to verify timestamps"),
        default="https://www.freetsa.org/index_en.php",
        required=True,
    )



# -*- coding: utf-8 -*-
from collective.timestamp.tests import TimestampIntegrationTestCase
from collective.timestamp.utils import get_timestamp_date_from_tsr_file
from collective.timestamp.utils import localize_utc_date
from collective.timestamp.utils import timestamp
from datetime import datetime
from datetime import timedelta
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.namedfile.file import NamedBlobFile
from unittest.mock import patch

import pytz


class TestUtils(TimestampIntegrationTestCase):

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.file_data = self.file.file.data

    def test_localize_utc_date(self):
        naive_datetime = datetime(2024, 9, 10, 12, 0, 0)
        localized_datetime = localize_utc_date(naive_datetime)
        expected_datetime = datetime(2024, 9, 10, 12, 0, 0, tzinfo=pytz.UTC)
        self.assertEqual(localized_datetime, expected_datetime)

    def test_timestamp(self):
        with patch("rfc3161ng.api.RemoteTimestamper.__call__") as mock_timestamper:
            mock_timestamper.return_value = self.tsr
            tsr, timestamp_date = timestamp(self.file_data, "http://freetsa.org/tsr")
        self.assertIsInstance(tsr, bytes)
        self.assertIsInstance(timestamp_date, datetime)
        self.assertEqual(
            timestamp_date, datetime(2025, 8, 25, 14, 49, 8, tzinfo=pytz.UTC)
        )

    def test_timestamp_retries(self):
        tsr, timestamp_date = timestamp(
            self.file_data,
            "https://httpbin.org/status/429",
            use_failover=True,
            failover_timestamping_service_urls=[
                "https://httpbin.org/status/429",
                "http://freetsa.org/tsr",
            ],
            max_retries=2,
            initial_backoff_seconds=0.1,
        )
        self.assertIsInstance(tsr, bytes)
        self.assertIsInstance(timestamp_date, datetime)
        self.assertAlmostEqual(
            timestamp_date, datetime.now(pytz.UTC), delta=timedelta(seconds=10)
        )

    def test_timestamp_raises_connection_error(self):
        with self.assertRaises(ConnectionError):
            timestamp(
                self.file_data,
                "https://httpbin.org/status/500",
                use_failover=True,
                failover_timestamping_service_urls=["https://httpbin.org/status/429"],
                max_retries=2,
                initial_backoff_seconds=0.01,
            )

    def test_get_timestamp_date_from_tsr_file(self):
        timestamp_date = get_timestamp_date_from_tsr_file(self.raw_tsr)
        self.assertEqual(
            timestamp_date, datetime(2025, 8, 25, 14, 49, 8, tzinfo=pytz.UTC)
        )



# -*- coding: utf-8 -*-

from collective.timestamp.behaviors.timestamp import ITimestampableDocument
from collective.timestamp.interfaces import ITimeStamper
from plone.indexer import indexer


@indexer(ITimestampableDocument)
def is_timestamped(obj):
    handler = ITimeStamper(obj)
    return handler.is_timestamped()



# -*- coding: utf-8 -*-
from collective.timestamp.browser.viewlet import TimestampViewlet
from collective.timestamp.interfaces import ITimeStamper
from collective.timestamp.tests import TimestampIntegrationTestCase
from datetime import datetime
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.namedfile.file import NamedBlobFile
from zope.component import getMultiAdapter
from zope.interface.interfaces import ComponentLookupError

import pytz


class TestInfos(TimestampIntegrationTestCase):

    def _timestamped_view(self):
        return getMultiAdapter(
            (self.timestamped_file, self.request), name="timestamp-info"
        )

    def _not_timestamped_view(self):
        return getMultiAdapter((self.file, self.request), name="timestamp-info")

    def test_viewlet(self):
        logout()
        viewlet = TimestampViewlet(self.document, self.request, None)
        with self.assertRaises(TypeError):
            viewlet.available()
        viewlet = TimestampViewlet(self.file, self.request, None)
        self.assertFalse(viewlet.available())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.assertFalse(viewlet.available())
        handler = ITimeStamper(self.file)
        handler.timestamp()
        self.assertTrue(viewlet.available())
        viewlet.update()
        html = viewlet.render()
        self.assertIn("svg", html)

    def test_view(self):
        logout()
        not_timestamped_view = self._not_timestamped_view()
        with self.assertRaises(ComponentLookupError):
            getMultiAdapter((self.document, self.request), name="timestamp")
        self.assertIn("This content is not timestamped.", not_timestamped_view())
        self.assertFalse(not_timestamped_view.is_timestamped())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.assertIn("This content is not timestamped.", not_timestamped_view())
        self.assertFalse(not_timestamped_view.is_timestamped())

        timestamped_view = self._timestamped_view()

        self.assertIn("Timestamped on", timestamped_view())
        self.assertEqual(
            timestamped_view.more_infos_url(), "http://documentation.timestamptest.com"
        )
        self.assertTrue(timestamped_view.is_timestamped())
        self.assertEqual(
            timestamped_view.timestamp_date(),
            datetime(2025, 8, 25, 14, 49, 8, tzinfo=pytz.UTC),
        )

    def test_timestamp_authority(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertDictEqual(not_timestamped_view.timestamp_authority(), {})

        timestamped_view = self._timestamped_view()
        timestamp_authority = timestamped_view.timestamp_authority()
        self.assertSetEqual(
            set(timestamp_authority.keys()),
            {
                "country_name",
                "locality_name",
                "organization_name",
                "organization_identifier",
                "common_name",
            },
        )

    def test_timestamp_precision(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertIsNone(not_timestamped_view.timestamp_precision())

        timestamped_view = self._timestamped_view()
        self.assertEqual(timestamped_view.timestamp_precision(), "1 second(s)")

    def test_timestamp_protocol(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertIsNone(not_timestamped_view.timestamp_protocol())

        timestamped_view = self._timestamped_view()
        self.assertEqual(timestamped_view.timestamp_protocol(), "RFC 3161")

    def test_timestamp_algorithm(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertIsNone(not_timestamped_view.timestamp_algorithm())

        timestamped_view = self._timestamped_view()
        self.assertEqual(timestamped_view.timestamp_algorithm(), "sha256")

    def test_timestamp_hash(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertIsNone(not_timestamped_view.timestamp_hash())

        timestamped_view = self._timestamped_view()
        self.assertIsNotNone(timestamped_view.timestamp_hash())
        self.assertEqual(
            timestamped_view.timestamp_hash(),
            "86f3c70fb6673cf303d2206db5f23c237b665d5df9d3e44efef5114845fc9f59",
        )

    def test_policy_oid_present(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertIsNone(not_timestamped_view.timestamp_policy_oid())

        timestamped_view = self._timestamped_view()
        oid = timestamped_view.timestamp_policy_oid()
        self.assertEqual("2.16.56.13.6.3.1.1000", oid)

    def test_serial_is_int_and_positive(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertIsNone(not_timestamped_view.timestamp_serial_number())

        timestamped_view = self._timestamped_view()
        serial = timestamped_view.timestamp_serial_number()
        self.assertEqual(serial, 7378481000937193731)

    def test_timestamp_file_and_tsr_links(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertIsNone(not_timestamped_view.timestamp_tsr_link())

        timestamped_view = self._timestamped_view()
        file_link = timestamped_view.timestamp_file_link()
        tsr_link = timestamped_view.timestamp_tsr_link()
        # File link
        self.assertIsNotNone(file_link)
        self.assertEqual(file_link["label"], "file.txt")
        self.assertIn("/@@download/file/file.txt", file_link["url"])
        # TSR link
        self.assertIsNotNone(tsr_link)
        self.assertTrue(tsr_link["label"].lower().endswith(".tsr"))
        self.assertIn("/@@download/timestamp/", tsr_link["url"])

    def test_certificate_chain_nonempty_and_order(self):
        not_timestamped_view = self._not_timestamped_view()
        self.assertListEqual(not_timestamped_view.certificate_chain(), [])

        timestamped_view = self._timestamped_view()
        chain = timestamped_view.certificate_chain()
        # At least the signer should be present
        self.assertGreaterEqual(len(chain), 1)

        # The first element is the signer; check that matches signer_certificate()
        signer = timestamped_view.signer_certificate()
        self.assertIsNotNone(signer)
        self.assertIs(chain[0], signer)

    def test_certificate_chain_info_shapes_and_dates(self):
        timestamped_view = self._timestamped_view()

        info = timestamped_view.certificate_chain_info()
        self.assertIsInstance(info, list)
        self.assertGreaterEqual(len(info), 1)

        first = info[0]
        for key in (
            "subject",
            "issuer",
            "serial_number",
            "fingerprint",
            "self_issued",
            "has_ski",
            "has_aki",
            "not_before",
            "not_after",
        ):
            self.assertIn(key, first)

        self.assertIsInstance(first["serial_number"], int)
        self.assertIsInstance(first["fingerprint"], str)
        self.assertRegex(
            first["fingerprint"], r"^[0-9a-f]{64}$"
        )  # sha256 hex by default
        self.assertIsInstance(first["self_issued"], bool)
        self.assertIsInstance(first["has_ski"], bool)
        self.assertIsInstance(first["has_aki"], bool)
        self.assertIsInstance(first["not_before"], datetime)
        self.assertIsInstance(first["not_after"], datetime)
        self.assertLess(first["not_before"], first["not_after"])

    def test_timestamp_signer_subject_mapping(self):
        timestamped_view = self._timestamped_view()
        subj = timestamped_view.timestamp_signer_subject()
        self.assertIsInstance(subj, dict)
        for k, v in subj.items():
            self.assertIn("label", v)
            self.assertIn("value", v)
            break  # one is enough to prove shape

    def test_certificates_in_token_basic(self):
        timestamped_view = self._timestamped_view()
        certs = timestamped_view.certificates_in_token()
        # There should be at least the signer certificate
        self.assertIsInstance(certs, list)
        self.assertGreaterEqual(len(certs), 1)



from collective.timestamp import logger
from datetime import datetime
from pyasn1.codec.der import decoder
from pyasn1.codec.der import encoder
from rfc3161ng import get_timestamp
from rfc3161ng import RemoteTimestamper
from rfc3161ng import TimeStampResp
from rfc3161ng import TimeStampToken

import pytz
import time


def localize_utc_date(date: datetime):
    tzinfo = pytz.timezone("UTC")
    return tzinfo.localize(date)


def get_timestamp_date(timestamp_token: TimeStampToken, localize=True):
    """
    Extract the timestamp date from a time stamp token.
    :param timestamp_token: The time stamp token to extract the date from.
    :param localize: Whether to localize the date to UTC (default is True).
    :return: The timestamp date, localized if specified.
    """
    timestamp_date = get_timestamp(timestamp_token)
    if localize:
        return localize_utc_date(timestamp_date)
    return timestamp_date


def get_timestamp_date_from_tsr_file(tsr_data: bytes, localize: bool = True):
    """Extracts the timestamp date from a tsr file"""
    tsr, _ = decoder.decode(tsr_data, asn1Spec=TimeStampResp())
    timestamp_token = tsr.time_stamp_token
    return get_timestamp_date(timestamp_token, localize=localize)


def timestamp(
    file_content: bytes,
    service_url: str,
    hashing_algorithm: str = "sha256",
    use_failover: bool = False,
    failover_timestamping_service_urls: list = None,
    max_retries: int = 0,
    initial_backoff_seconds: float = 0.5,
    localize_date: bool = True,
):
    """
    Generate a timestamp for the given file_content using a remote timestamping service.
    It attempts to obtain a timestamp from the (primary) service_url, and if it fails,
    it will switch to one of the failover_timestamping_service_urls if configured.
    It can also use exponential backoff for retries.
    :param file_content: The content of the file to be timestamped.
    :param service_url: The primary timestamping service URL.
    :param hashing_algorithm: The hashing algorithm to use (default is "sha256").
    :param use_failover: Whether to use failover timestamping services if the primary service fails.
    :param failover_timestamping_service_urls: A list of failover timestamping service URLs.
    :param max_retries: The maximum number of retries for timestamping attempts.
    :param initial_backoff_seconds: The initial backoff time in seconds for retrying failed attempts.
    :raises ConnectionError: If all attempts to obtain a timestamp fail.
    :return: A tuple containing the encoded timestamp response and the localized timestamp date.
    """
    success = False
    tsr = None
    # Make a copy to avoid modifying the original list
    failover_urls = (
        failover_timestamping_service_urls.copy()
        if failover_timestamping_service_urls
        else []
    )
    while not success:
        retry_count = 0
        backoff_seconds = initial_backoff_seconds

        # Try main service URL with exponential backoff
        while retry_count <= max_retries:
            try:
                timestamper = RemoteTimestamper(
                    service_url,
                    certificate=b"",
                    hashname=hashing_algorithm,
                )
                tsr = timestamper(
                    data=file_content,
                    include_tsa_certificate=True,
                    return_tsr=True,
                )
                success = True
                break
            except Exception as e:
                retry_count += 1
                logger.error(f"Timestamping attempt {retry_count} failed: {e}")
                if retry_count <= max_retries:
                    logger.info(f"Retrying in {backoff_seconds} seconds...")
                    time.sleep(backoff_seconds)
                    backoff_seconds *= 2
                else:
                    logger.error(
                        f"Reached max retries ({max_retries}) for URL: {service_url}"
                    )

        if success or not use_failover:
            # If we succeeded or not using failover, we can exit the loop
            break

        if use_failover:
            if not failover_urls:
                logger.error("No failover URLs available, cannot proceed.")
                break
            service_url = failover_urls.pop(0)
            logger.info(f"Switching to failover URL: {service_url}")
    if not success:
        raise ConnectionError("Failed to obtain a timestamp.")

    timestamp_token = tsr.time_stamp_token
    timestamp_date = get_timestamp_date(timestamp_token, localize=localize_date)
    return encoder.encode(tsr), timestamp_date



from collective.timestamp import logger
from collective.timestamp.behaviors.timestamp import ITimestampableDocument
from collective.timestamp.interfaces import ITimeStamper
from OFS.SimpleItem import SimpleItem
from plone import api
from plone.app.contentrules import PloneMessageFactory as _
from plone.app.contentrules.browser.formhelper import NullAddForm
from plone.base.utils import pretty_title_or_id
from plone.contentrules.rule.interfaces import IExecutable
from plone.contentrules.rule.interfaces import IRuleElementData
from rfc3161ng import TimestampingError
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


class ITimestampAction(Interface):
    """Interface for the configurable aspects of a timestamp action."""


@implementer(ITimestampAction, IRuleElementData)
class TimestampAction(SimpleItem):
    """The actual persistent implementation of the action element."""

    element = "collective.timestamp.actions.Timestamp"
    summary = _("Timestamp object")


@adapter(Interface, ITimestampAction, Interface)
@implementer(IExecutable)
class TimestampActionExecutor:
    """The executor for this action."""

    def __init__(self, context, element, event):
        self.context = context
        self.element = element
        self.event = event

    def __call__(self):
        obj = self.event.object
        if not obj:
            return False

        if not ITimestampableDocument.providedBy(obj):
            return False
        handler = ITimeStamper(obj)
        if not handler.is_timestampable():
            return False
        try:
            handler.timestamp()
        except TimestampingError as e:
            self.error(obj, str(e))
            logger.error(f"Timestamp rule failed for {obj.absolute_url()} : {str(e)}")
            return False

        logger.info(f"Timestamp generated for {obj.absolute_url()}")
        request = getattr(self.context, "REQUEST", None)
        if request is not None:
            title = pretty_title_or_id(obj, obj)
            message = _(
                "Timestamp file has been successfully generated and saved for ${name}",
                mapping={"name": title},
            )
            api.portal.show_message(message, request)
        return True

    def error(self, obj, error):
        request = getattr(self.context, "REQUEST", None)
        if request is not None:
            title = pretty_title_or_id(obj, obj)
            message = _(
                "Unable to timestamp ${name}: ${error}",
                mapping={"name": title, "error": error},
            )
            api.portal.show_message(message, request, type="error")


class TimestampAddForm(NullAddForm):
    """A degenerate "add form" for timestamp actions."""

    def create(self):
        return TimestampAction()



# -*- coding: utf-8 -*-

from collective.timestamp.interfaces import ITimeStamper
from plone.app.layout.viewlets import common
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile


class TimestampViewlet(common.ViewletBase):
    index = ViewPageTemplateFile("templates/viewlet.pt")

    def available(self):
        handler = ITimeStamper(self.context)
        return handler.is_timestamped()
